@@ -3,7 +3,7 @@
 
 # EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
 #
-# Copyright (C) 2018-2022 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
+# Copyright (C) 2018-2023 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
 # (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
 # Stephane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
 #
@@ -114,6 +114,8 @@ def get_enpt_argparser():
     add('--mode_ac', type=str, default=None, nargs='?',
         help="3 modes to determine which atmospheric correction is applied at which surfaces (default: land): "
              "('land', water', 'combined')")
+    add('--polymer_additional_results', type=_str2bool, default=True, nargs='?', const=True,
+        help="Enable the generation of additional results when running ACwater/POLYMER (default: True)")
     add('--auto_download_ecmwf', type=_str2bool, default=True, nargs='?', const=True,
         help='Automatically download ECMWF AUX data when running Polymer atmospheric correction for water surfaces')
     add('--scale_factor_boa_ref', type=int, default=10000,
