@@ -2,7 +2,7 @@
 
 # EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
 #
-# Copyright (C) 2018-2022 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
+# Copyright (C) 2018-2023 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
 # (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
 # Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
 #
@@ -94,9 +94,9 @@ class EnMAP_Metadata_L2A_MapGeo(object):
         file_ext_l1b = os.path.splitext(meta_l1b.vnir.filename_data)[1]
         file_ext_l2a = \
             '.TIF' if self.cfg.output_format == 'GTiff' else \
-            '.bsq' if self.cfg.output_interleave == 'band' else \
-            '.bil' if self.cfg.output_interleave == 'line' else \
-            '.bip'
+            '.BSQ' if self.cfg.output_interleave == 'band' else \
+            '.BIL' if self.cfg.output_interleave == 'line' else \
+            '.BIP'
 
         def convert_fn(fn):
             return fn.replace('L1B-', 'L2A-').replace(file_ext_l1b, file_ext_l2a)
