@@ -2,7 +2,7 @@
 
 # EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
 #
-# Copyright (C) 2018-2022 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
+# Copyright (C) 2018-2023 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
 # (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
 # Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
 #
@@ -85,6 +85,7 @@ enpt_schema_input = dict(
                     polymer_root=dict(type='string', required=False),
                     enable_ac=dict(type='boolean', required=False),
                     mode_ac=dict(type='string', required=False, allowed=['land', 'water', 'combined']),
+                    polymer_additional_results=dict(type='boolean', required=False),
                     auto_download_ecmwf=dict(type='boolean', required=False),
                     scale_factor_boa_ref=dict(type='integer', required=False, min=1),
                     threads=dict(type='integer', required=False),
@@ -159,6 +160,7 @@ parameter_mapping = dict(
     polymer_root=('processors', 'atmospheric_correction', 'polymer_root'),
     enable_ac=('processors', 'atmospheric_correction', 'enable_ac'),
     mode_ac=('processors', 'atmospheric_correction', 'mode_ac'),
+    polymer_additional_results=('processors', 'atmospheric_correction', 'polymer_additional_results'),
     auto_download_ecmwf=('processors', 'atmospheric_correction', 'auto_download_ecmwf'),
     scale_factor_boa_ref=('processors', 'atmospheric_correction', 'scale_factor_boa_ref'),
     threads=('processors', 'atmospheric_correction', 'threads'),
