@@ -2,7 +2,7 @@
 
 # EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
 #
-# Copyright (C) 2018-2022 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
+# Copyright (C) 2018-2023 Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
 # (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
 # Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
 #
@@ -34,22 +34,11 @@ Performs the atmospheric correction of EnMAP L1B data.
 import pprint
 import numpy as np
 from multiprocessing import cpu_count
-from typing import Optional
 from logging import Logger
-from warnings import warn
 
+from packaging.version import parse as parse_version
 from sicor.sicor_enmap import sicor_ac_enmap
 from sicor.options import get_options as get_ac_options
-try:
-    from acwater.acwater import polymer_ac_enmap
-except ImportError as e:
-    try:
-        import acwater as _acwater  # noqa: F401
-        warn(f'ACwater is importable but the following error occurred when importing polymer_ac_enmap: {e.msg}')
-    except ImportError as e:
-        warn(f'ACwater cannot be imported. The following error occurred:  {e.msg}')
-    finally:
-        polymer_ac_enmap: Optional[callable] = None
 
 from ...model.images import EnMAPL1Product_SensorGeo
 from ...options.config import EnPTConfig
@@ -64,7 +53,6 @@ class AtmosphericCorrector(object):
     def __init__(self, config: EnPTConfig = None):
         """Create an instance of AtmosphericCorrector."""
         self.cfg = config
-        self.is_polymer_installed = polymer_ac_enmap is not None
 
     def _get_sicor_options(self, enmap_ImageL1: EnMAPL1Product_SensorGeo, land_only=False) -> dict:
         """Get a dictionary containing the SICOR options.
@@ -94,8 +82,15 @@ class AtmosphericCorrector(object):
             options["retrieval"]["land_only"] = land_only
 
             # disable first guess water vapor retrieval for now
-            # options["retrieval"]["state_vector"]["water_vapor"]["use_prior_mean"] = True
-            # options["retrieval"]["state_vector"]["water_vapor"]["prior_mean"] = 2.5  # = already the default
+            options["retrieval"]["state_vector"]["water_vapor"]["use_prior_mean"] = True
+            options["retrieval"]["state_vector"]["water_vapor"]["prior_mean"] = \
+                enmap_ImageL1.meta.water_vapour  # = default = 2.5
+
+            # disable first guess liquid water retrieval for now
+            options["retrieval"]["state_vector"]["liquid_water"]["use_prior_mean"] = True
+
+            # disable first guess ice retrieval for now
+            options["retrieval"]["state_vector"]["ice"]["use_prior_mean"] = True
 
         except FileNotFoundError:
             raise FileNotFoundError(f'Could not locate options file for atmospheric correction at {path_opts}')
@@ -105,6 +100,41 @@ class AtmosphericCorrector(object):
 
         return options
 
+    def _is_acwater_operable(self, logger: Logger):
+        """Return True if ACWater/Polymer is operable, else raise a warning and return False."""
+        try:
+            import acwater as _acwater  # noqa: F401
+
+            if parse_version(_acwater.__version__) < parse_version('0.3.0'):
+                if self.cfg.mode_ac in ['water', 'combined']:
+                    logger.warning(f"The installed version of ACwater (v{_acwater.__version__}) is too old. "
+                                   f"At least version 0.3.0 is required. Instead of ACwater, SICOR is applied to water "
+                                   f"surfaces as a workaround.")
+
+                return False
+
+        except ImportError as e:
+            if self.cfg.mode_ac in ['water', 'combined']:
+                logger.warning(f"The atmospheric correction mode was set to '{self.cfg.mode_ac}' but "
+                               f"ACwater cannot be imported (Error was: {e.msg}). "
+                               f"As a fallback, SICOR is applied to water surfaces instead.")
+            return False
+
+        try:
+            from acwater.acwater import polymer_ac_enmap as _polymer_ac_enmap
+            if not _polymer_ac_enmap:
+                logger.warning("Polymer is not callable. "
+                               "As a fallback, SICOR is applied to water surfaces instead.")
+                return False
+        except ImportError as e:
+            if self.cfg.mode_ac in ['water', 'combined']:
+                logger.warning(f"The atmospheric correction mode was set to '{self.cfg.mode_ac}' but "
+                               f"Polymer cannot be imported (Error was: {e.msg}). "
+                               f"As a fallback, SICOR is applied to water surfaces instead.")
+            return False
+
+        return True
+
     def _run_AC__land_mode(self,
                            enmap_ImageL1: EnMAPL1Product_SensorGeo
                            ) -> (np.ndarray, np.ndarray, dict):
@@ -141,6 +171,8 @@ class AtmosphericCorrector(object):
             - Land surfaces are NOT included in the EnMAP L2A product.
             - The output radiometric unit for water surfaces is 'water leaving reflectance'.
         """
+        from acwater.acwater import polymer_ac_enmap
+
         if 1 in enmap_ImageL1.vnir.mask_landwater[:]:
             enmap_ImageL1.logger.info(
                 "Running atmospheric correction in 'water' mode, i.e., ACWater/Polymer is applied to water "
@@ -148,11 +180,13 @@ class AtmosphericCorrector(object):
 
         # run ACWater/Polymer for water surfaces only
         # NOTE: polymer_ac_enmap() returns masked (nan) values for land
+        #       - res: a dictionary containing retrieval maps with several additional retrieval measures
+        #              -> chla, bitmask, logfb, Rnir, Rgli
         try:
-            wl_ref_vnir, wl_ref_swir = \
+            wl_ref_vnir, wl_ref_swir, water_additional_results = \
                 polymer_ac_enmap(enmap_l1b=enmap_ImageL1,
                                  config=self.cfg,
-                                 detector='merge')
+                                 detector='vnir')
         except:  # noqa
             enmap_ImageL1.logger.error(
                 "The atmospheric correction for water surfaces based on ACwater/Polymer failed (issue tracker at "
@@ -162,7 +196,7 @@ class AtmosphericCorrector(object):
             )
             raise
 
-        return wl_ref_vnir, wl_ref_swir
+        return wl_ref_vnir, wl_ref_swir, water_additional_results
 
     def _run_AC__combined_mode(self,
                                enmap_ImageL1: EnMAPL1Product_SensorGeo
@@ -175,6 +209,8 @@ class AtmosphericCorrector(object):
                 - 'water leaving reflectance' for water surfaces
             - There might be visible edge effects, e.g., at coastlines.
         """
+        from acwater.acwater import polymer_ac_enmap
+
         only = 'water' if 1 not in enmap_ImageL1.vnir.mask_landwater[:] else 'land'
         if 1 not in enmap_ImageL1.vnir.mask_landwater[:] or \
            2 not in enmap_ImageL1.vnir.mask_landwater[:]:
@@ -191,11 +227,13 @@ class AtmosphericCorrector(object):
 
         # run ACWater/Polymer for water surfaces only
         # NOTE: polymer_ac_enmap() returns masked (nan) values for land
+        #       - res: a dictionary containing retrieval maps with several additional retrieval measures
+        #              -> chla, bitmask, logfb, Rnir, Rgli
         try:
-            wl_ref_vnir_water, wl_ref_swir_water = \
+            wl_ref_vnir_water, wl_ref_swir_water, water_additional_results = \
                 polymer_ac_enmap(enmap_l1b=enmap_ImageL1,
                                  config=self.cfg,
-                                 detector='merge')
+                                 detector='vnir')
         except:  # noqa
             enmap_ImageL1.logger.error(
                 "The atmospheric correction for water surfaces based on ACwater/Polymer failed (issue tracker at "
@@ -205,15 +243,13 @@ class AtmosphericCorrector(object):
             )
             raise
 
-        # use mask value 2 for replacing water corrected pixels
-        wlboa_ref_vnir = np.where((enmap_ImageL1.vnir.mask_landwater[:] == 2)[:, :, None],
-                                  wl_ref_vnir_water,
-                                  boa_ref_vnir_land)
-        wlboa_ref_swir = np.where((enmap_ImageL1.swir.mask_landwater[:] == 2)[:, :, None],
-                                  wl_ref_swir_water,
-                                  boa_ref_swir_land)
+        # Overwrite the SICOR output at water positions with the output from ACwater/Polymer
+        # -> output contains Water-leaving-reflectance over water and BOA-reflectance over land
+        water_mask_vnir_3D = (enmap_ImageL1.vnir.mask_landwater[:] == 2)[:, :, None]  # 2 = water
+        wlboa_ref_vnir = np.where(water_mask_vnir_3D, wl_ref_vnir_water, boa_ref_vnir_land)
+        wlboa_ref_swir = np.where(water_mask_vnir_3D, wl_ref_swir_water, boa_ref_swir_land)
 
-        return wlboa_ref_vnir, wlboa_ref_swir, land_additional_results
+        return wlboa_ref_vnir, wlboa_ref_swir, water_additional_results, land_additional_results
 
     @staticmethod
     def _validate_AC_results(reflectance_vnir: np.ndarray,
@@ -244,23 +280,23 @@ class AtmosphericCorrector(object):
             f"Starting atmospheric correction for VNIR and SWIR detector in '{self.cfg.mode_ac}' mode. "
             f"Source radiometric unit code is '{enmap_ImageL1.meta.vnir.unitcode}'.")
 
-        # run the AC
-        if not self.is_polymer_installed and self.cfg.mode_ac in ['water', 'combined']:
-            # use SICOR as fallback AC for water surfaces if ACWater/Polymer is not installed
-            enmap_ImageL1.logger.warning(
-                f"The atmospheric correction mode was set to '{self.cfg.mode_ac}' but the packages ACWater/Polymer are "
-                f"missing. SICOR has to be used as fallback algorithm for water surfaces.")
+        # set initial values for land_additional_results and water_additional_results
+        land_additional_results = None
+        water_additional_results = None
 
+        # run the AC
+        if self.cfg.mode_ac in ['water', 'combined'] and not self._is_acwater_operable(enmap_ImageL1.logger):
+            # use SICOR as fallback AC for water surfaces if ACWater/Polymer is not installed
             reflectance_vnir, reflectance_swir, land_additional_results = \
                 self._run_AC__land_mode(enmap_ImageL1)
 
         else:
             if self.cfg.mode_ac == 'combined':
-                reflectance_vnir, reflectance_swir, land_additional_results = \
+                reflectance_vnir, reflectance_swir, water_additional_results, land_additional_results = \
                     self._run_AC__combined_mode(enmap_ImageL1)
 
             elif self.cfg.mode_ac == 'water':
-                reflectance_vnir, reflectance_swir = \
+                reflectance_vnir, reflectance_swir, water_additional_results = \
                     self._run_AC__water_mode(enmap_ImageL1)
 
             elif self.cfg.mode_ac == 'land':
@@ -289,4 +325,25 @@ class AtmosphericCorrector(object):
         # FIXME: Consider to also join SICOR's land_additional_results
         #  (contains three phases of water maps and several retrieval uncertainty measures)
 
+        # join additional results from ACwater/Polymer
+        if water_additional_results and self.cfg.polymer_additional_results:
+
+            water_mask = enmap_ImageL1.vnir.mask_landwater[:] == 2
+            for k in water_additional_results.keys():
+                if k == 'polymer_bitmask':
+                    # the bitmask already explicitly indicates land pixels with "1"
+                    continue
+                else:
+                    v = water_additional_results[k]
+                    v[~water_mask] = -9999
+                    v[np.isnan(v)] = -9999
+
+                    water_additional_results[k] = v
+
+            enmap_ImageL1.vnir.polymer_logchl = water_additional_results['polymer_logchl']
+            enmap_ImageL1.vnir.polymer_logfb = water_additional_results['polymer_logfb']
+            enmap_ImageL1.vnir.polymer_rgli = water_additional_results['polymer_rgli']
+            enmap_ImageL1.vnir.polymer_rnir = water_additional_results['polymer_rnir']
+            enmap_ImageL1.vnir.polymer_bitmask = water_additional_results['polymer_bitmask']
+
         return enmap_ImageL1
